/*
 Program Ŀ
  Application: Freshwinized Error System                                  
    File Name: FWERS501.PRG                                               
 Date created: 12-20-92                                                   
 Time created: 02:58:41am                                                 
                                                                          
 Based upon the original work of Tim Palmquist                            
*/
/* 
 Notes:
 YOU MUST USE FRESHWIN with this!
 NOT FOR USE WITH CLIPPER SUMMER '87, 5.00, OR EARLIER VERSIONS!
 
 To use this error system with an application, compile this file
 with options "/n /w /m", then link ERS501.OBJ in with your program.

 Along with reporting all available information about this error and
 allowing the user to decide what action to take, this system also allows
 an error-tracking mechanism to be included, if desired.  All error
 information will be written to a text file, and all memory variables will
 be saved to a .MEM file.  If you are a consultant (like myself), it may
 be helpful to track errors which your client has not reported.  If you do
 not wish to track errors in this manner, just comment out the "#define"
 lines below for MAKE_LOG and MAKE_MEM.  (Please also note that if Clipper
 cannot create the error log file, the error system will appear to operate
 in its usual manner, without actually tracking the errors.)

 This error system has been tested and verified as functioning properly,
 with one exception.  If pending PROMPTs exist (before a MENU TO
 command has been issued) when the error occurs, the pending prompts
 are included in the error-action menu, resulting in an internal error.
 Any suggestions on ways to avoid this problem would be appreciated.

 Please note also that Clipper bypasses the error system for some
 severe errors.  I am not aware of any way to trap these severe errors,
 other that hooking into INT 24 and I unfortunately haven't got time to
 rig'n'write that in assembler.

 For those who are curious, the difference between version 1.00 of this
 error system and this current version is primarily compatibility with
 Clipper 5.01.  Version 1.00 was compatible with Clipper 5.0, but not
 with 5.01; this version is compatible with Clipper 5.01, but not 5.0.
 However, a few additional changes were made:  The severity level is
 now reported, and special handling is provided for printer errors, along
 with zero division and network append lock errors.  The system still
 makes no use of the error object's "cargo"; if you have any ideas for
 using this capability, please contact me.

  */

   #include "error.ch"
   #include "set.ch"
   #include "fw.ch"

   //#define CRLF Chr(13)+Chr(10)

   /* comment out below line to turn off error log-file creation */
   //#define MAKE_LOG 'CLIP'

   /* comment out below line to turn off error mem-file creation */
   //#define MAKE_MEM 'CLIP'

   Static nLogFile

/* Procedure Ŀ
  Description: Executes automatically when program starts                 
*/
Proc ErrorSys()

   ErrorBlock( {|oError| ErrorMenu(oError)} )

   return

/*
 Function Ŀ
  Description:                                                            
                                                                          
   |-Error: BASE/9999---------------------------|                         
   | Code 99999: DESCRIPTION______  (retry 999) |                         
   | Dos error 999: DESCRIPTION________________ |                         
   | Operation attempted: ____________          |                         
   | Arguments: XXXXXXXXXX & XXXXXXXXXXXXXX     |                         
   | Filename: XXXXXXXXXXXXXXXXX                |                         
   | Called from: XXXXXXXXXX (111)              |                         
   |              XXXXXXXXXX (XXX)              |                         
   |              XXXXXXXXXX (XXX)              |                         
   | Press <Shift><PrtSc> to print this screen. |                         
   | Then notify the developer of this error.   |                         
   |     |--------------------------------|     |                         
   |     | Retry Ignore Fail Screen Abort |     |                         
   |_____|________________________________|_____|                         
                                                                          
   NOTE: none of the above lines will display unless there is             
         an appropriate value to display.  (For example, the              
         "Filename:" line will not display unless this error              
         relates to a file.) DosErr[] is an array containing              
         error messages corresponding to some of the DOS error            
         numbers.  Messages may be edited or changed to null              
         strings, if desired.                                             
                                                                          
 Date created: 12-20-92                                                   
 Time created: 03:02:03am                                                 
 */
static Func ErrorMenu(oer)

   memvar errscreen
   local nerrlines,nactcount,cchoices,nerrtop,nchoice,cchoice,cerrscr,nintense,;
         cdevice,ccolor,bolderr,cmsg,cerrwrnmsg
   local doserr:={"Invalid function number","File not found","Path not found",;
         "Too many open files","Access denied","Invalid handle","Memory",;
         "Insufficient memory","Memory","Invalid environment","Invalid format",;
         "Invalid access code","Invalid data","","Invalid drive",;
         "Cannot remove directory","Not same device","No more files","Write-protect",;
         "Unknown unit","Drive not ready","Unknown command","Data error (CRC)",;
         "Bad request","Seek error","Unknown media type","Sector not found",;
         "Printer out of paper","Write fault","Read fault","General failure",;
         "Sharing violation","Lock violation","Invalid disk change","FCB unavailable",;
         "Sharing buffer overflow","","","","","","","","","","","","","",;
         "Network request not supported","Remote not listening",;
         "Duplicate name on network","Network name not found","Network busy",;
         "Network device no longer exists","Network BIOS command limit exceeded",;
         "Network adapter hardware error","Incorrect response from network",;
         "Unexpected network error","Incompatible remote adapter","Print queue full",;
         "Not enough space for print file","Print file deleted","Network name deleted",;
         "Access denied","Network device type incorrect","Network name not found",;
         "Network name limit exceeded","Network BIOS session limit exceeded",;
         "Temporarily paused","Network request not accepted",;
         "Print or disk redirection paused","","","","","","","","File already exists",;
         "","Cannot make directory entry","Fail on INT 24H","Too many redirections",;
         "Duplicate redirection","Invalid password","Invalid parameter",;
         "Network device fault"}

   //*  by default, division by zero yields zero
   if (oEr:genCode == EG_ZERODIV)
       return (0)
   endif

   //*  for network open error, set NetErr() and subsystem default
   if ((oEr:genCode == EG_OPEN) .and. oEr:osCode == 32 .and. (oEr:canDefault))
        neterr(.t.)
        return (.f.)
   endif

   //*  check for lock error during APPEND BLANK
   if (oEr:genCode == EG_APPENDLOCK .and. oEr:canDefault)
       neterr(.t.)
       return (.f.)
   endif

   //*  If error occurs within this error-handler, handle with DuplError
   bOldErr:=ErrorBlock( {|oError| DuplError(oError)} )

   nlogfile=-1
   if (oer:gencode != EG_PRINT)
   #ifdef MAKE_LOG
   nerrlines:=0
   while (file(MAKE_LOG+ltrim(str(nerrlines))+'.ERR') .and. (nerrlines<10000))
         nerrlines++
   enddo

   if (nerrlines<10000)
      *   open error log file (if possible)
      nlogfile=fcreate(MAKE_LOG+ltrim(str(nerrlines))+'.ERR')
   endif
#endif

#ifdef MAKE_MEM
  nErrLines:=0
  Do While (File(MAKE_MEM+LTrim(Str(nErrLines))+'.MEM') .And. (nErrLines<10000))
     nErrLines++
  EndDo
  If (nErrLines<10000)
    Private ErrScreen
    Save Screen To ErrScreen
    Save To (MAKE_MEM+LTrim(Str(nErrLines))+'.MEM')
  EndIf
#endif
EndIf

cChoice:=' '
while cchoice!='A'
  /*  save original color and device setting,
     then set to display on screen with red background
     NOTE: IF YOU ARE USING THIS SYSTEM WITH A MONOCHROME MONITOR,
     CHANGE THE COLOR SETTING BELOW TO A MORE APPROPRIATE COMBINATION */
  cColor=SetColor('W/R+')
  cDevice=Set(_SET_DEVICE)
  Set Device To Screen

  /* count number of activations which we'll need to be displayed */
#ifdef MAKE_LOG
  nerrlines:=5
#else
  if (oer:gencode == EG_PRINT)
    nerrlines:=5
  else
    nerrlines:=6
  endif
#endif
  If (oEr:genCode != EG_PRINT)
    nActCount:=2
    Do While (.Not. (ProcName(nActCount)==""))
       nErrLines++
       nActCount++
    EndDo
    nErrLines+=IIf(oEr:operation=="",0,1)
    If (ValType(oEr:args)=='A')
       If (Len(oEr:args)>0)
          nErrLines++
       EndIf
    EndIf
    nErrLines+=IIf(oEr:filename=="",0,1)
    nErrLines+=IIf(oEr:osCode!=0,1,0)
  EndIf
  nErrTop=IIf(nErrLines<24,(24-nErrLines)/2,0)

  cErrScr=SaveScreen(nErrTop,17,nErrTop+nErrLines+2,62+2)
  @ nErrTop,17 Clear To nErrTop+nErrLines,62

  if (oEr:genCode == EG_PRINT)
   cErrWrnMsg='Printer error'
  Else
    If (oEr:severity==ES_CATASTROPHIC)
   cErrWrnMsg='Severe error'
    ElseIf (oEr:severity==ES_ERROR)
   cErrWrnMsg='Error'
    ElseIf (oEr:severity==ES_WARNING)
   cErrWrnMsg='Warning'
    Else
   cErrWrnMsg='Message'
    EndIf
#ifdef MAKE_LOG
    erlogwr(CRLF+'*** '+upper(cerrwrnmsg)+' ON '+dtoc(date())+' AT '+;
   Time()+' ***'+CRLF)
#endif
  endif
  setcolor('gr+/r')
  bOpen(nErrTop,17,nErrTop+nErrLines,62 ,"w+/r",B_EXPLODERT(10),B_SHADOW(,S_HALF),B_THICK)
  if (!empty(oEr:subSystem))
      cMsg=AllTrim(oEr:subsystem)+'/'+LTrim(Str(oEr:subCode))
      @ nErrTop,18 Say ' '+cErrWrnMsg+': '+cMsg+' '
#ifdef MAKE_LOG
      if (oEr:genCode != EG_PRINT)
         ErLogWr(cMsg+CRLF)
      endif
#endif
   endif
  if (oEr:genCode != EG_PRINT)
     setcolor('w/r')
     cMsg='Code '+LTrim(Str(oEr:genCode))+': '+oEr:description
     @ Row()+1,19 Say PadR(cMsg,60-Col())
#ifdef MAKE_LOG
    ErLogWr(cMsg)
#endif
    If (oEr:tries>1)
   cMsg='(retry '+LTrim(Str(oEr:tries))+')'
   @ Row(),48 Say '  '+cMsg
#ifdef MAKE_LOG
        ErLogWr(cMsg)
#endif
    EndIf
#ifdef MAKE_LOG
    ErLogWr(CRLF)
#endif
    If (oEr:osCode!=0)
   cMsg='Dos error '+LTrim(Str(oEr:osCode))
   @ Row()+1,19 Say cMsg+': '
   If ((oEr:oscode<=Len(DosErr)) .And. (oEr:oscode>0))
      @ Row(),Col() Say PadR(DosErr[oEr:osCode],60-Col())
   EndIf
#ifdef MAKE_LOG
        ErLogWr(cMsg+CRLF)
#endif
    EndIf
    If (.Not. (oEr:operation==""))
        cMsg='Operation attempted: '+oEr:operation
        @ Row()+1,19 Say PadR(cMsg,60-Col())
#ifdef MAKE_LOG
        ErLogWr(cMsg+CRLF)
#endif
    EndIf
    If (ValType(oEr:args)=='A')
      If (Len(oEr:args)>0)
   cMsg='Arguments: '+ErrDispArg(oEr:args[1])
   @ Row()+1,19 Say Left(cMsg,50-Col())
#ifdef MAKE_LOG
        ErLogWr(cMsg)
#endif
   If (Len(oEr:args)>1)
      cMsg='& '+ErrDispArg(oEr:args[2])
      @ Row(),Col() Say Left(cMsg,60-Col())
#ifdef MAKE_LOG
                ErLogWr(cMsg)
#endif
        EndIf
#ifdef MAKE_LOG
        ErLogWr(CRLF)
#endif
      EndIf
    EndIf
    If (.Not. (oEr:filename==""))
        cMsg='Filename: '+oEr:filename
        @ Row()+1,19 Say PadR(cMsg,60-Col())
#ifdef MAKE_LOG
        ErLogWr(cMsg+CRLF)
#endif
    endIf
    @ Row()+1,19 Say 'Called from: '
    nActCount:=2
    do while (!(procname(nactcount)==""))
        cmsg=procname(nactcount)+' ('+ltrim(str(procline(nactcount)))+')'
        If (row()<20)
                @ row(),32 say cMsg
                @ row()+1,col() Say ''
        endif
        #ifdef MAKE_LOG
        ErLogWr(cMsg+CRLF)
        #endif
        nActCount++
    enddo
    #ifdef MAKE_LOG
    @ Row(),19   Say 'Please notify Programmer of this error.'
    #else
    @ Row(),19   Say 'Press <Shift><PrtSc> to print this screen.'
    @ Row()+1,19 Say 'Then notify Programmer of this error.'
    #endif
  else
    @ Row()+1,19 Say 'Please check the printer to be sure that'
    @ Row()+1,19 Say 'it is turned on and ready to print.'
  EndIf
  @ Row()+2,19 Say ''
  nIntense=Set(_SET_INTENSITY,.T.)
  cChoices:=""
//  setcolor('bg/b,bg+/bg,,,bg/b')
  If (oEr:canRetry)
        cChoices+="R"
        @ Row(),25 Prompt 'Retry'
  Else
        @ Row(),25 Say 'Retry'
  EndIf
  If (oEr:canDefault)
        cChoices+="I"
        @ Row(),31 Prompt 'Ignore'
  Else
        @ Row(),31 Say 'Ignore'
  EndIf

  @ Row(),38 Prompt 'Fail'
  @ Row(),43 Prompt 'Screen'
  @ Row(),50 Prompt 'Abort'
  cChoices+='FSA'
  setcolor('gr+/r')
  @ Row()-1,23 Say   'Ŀ'
  @ Row()+1,23 Say   ''
  @ Row(),56 Say ''
  
  Menu To nChoice in window
  setcolor('bg/b,bg+/bg,,,bg/b')
  Set(_SET_INTENSITY,nIntense)
  Set(_SET_DEVICE,cDevice)
  SetColor(cColor)
  If (nChoice==0)
        * ESC re-views screen.
        cChoice='S'
  Else
        cChoice=SubStr(cChoices,nChoice,1)
  EndIf
#ifdef MAKE_LOG
  ErLogWr('User choice = '+cChoice+CRLF)
#endif
  RestScreen(nErrTop,17,nErrTop+nErrLines+2,62+2,cErrScr)
  If cChoice=='R'
#ifdef MAKE_LOG
   /* close error log file */
   FClose(nLogFile)
#endif
   ErrorBlock(bOldErr)
        Return .T.
  ElseIf cChoice=='I'
#ifdef MAKE_LOG
   /* close error log file */
   FClose(nLogFile)
#endif
   ErrorBlock(bOldErr)
        Return .F.
  ElseIf cChoice=='F'
#ifdef MAKE_LOG
   /* close error log file */
   FClose(nLogFile)
#endif
   ErrorBlock(bOldErr)
        Break
  ElseIf cChoice=='S'
        Inkey(0)
        Loop
  EndIf
EndDo cChoice!='A' // loop only happens if 'Screen' is selected.
#ifdef MAKE_LOG
/* close error log file */
FClose(nLogFile)
#endif

SetCursor(1) // don't leave them in DOS without a cursor!
ErrorLevel(1)
setcolor('w/n')
v_CloseAll()
dbcloseall()

      palette(1,1) // restore standard palette colors
      palette(2,2) //
      palette(5,5) //
      palette(6,6) //
setcolor("w+/n")
cls
Quit
Return .F.  //superfluous RETURN required to compile without warning

/*
 Function Ŀ
 */
Static Function ErrDispArg(arg)

   local cstr
   if valtype(arg)=='A'
           cstr:='(Array) '
   elseif valtype(arg)=='B'
           cstr:='(Block) '
   elseif valtype(arg)=='C'
           cstr:='(C) "'+arg+'" '
   elseif valtype(arg)=='D'
           cstr:='(D) '+dtoc(arg)+' '
   elseif valtype(arg)=='L'
           cstr:='(L) .'+iif(arg,'T','F')+'. '
   elseif valtype(arg)=='M'
           cstr:='(M) "'+arg+'" '
   elseif valtype(arg)=='N'
           cstr:='(N) '+alltrim(str(arg))+' '
   elseif valtype(arg)=='O'
           cstr:='(Object) '
   else
           cstr:='(NIL) '
   endif

   return cstr

/*
 Function Ŀ
 */
Static Function ErLogWr(cStr)

   if (nlogfile>-1)
       fwrite(nlogfile,cstr)
   endif

   return cstr

/*
 Function Ŀ
  Description: For errors that occur WITHIN the ErrorMenu error system    
 */
Static Function DuplError(oEr)

   ? "Error ("+Trim(oEr:description)+") at ",Trim(ProcName(2))+' ('+LTrim(Str(ProcLine(2)))+')'
   errorlevel(1)
   setcolor('w/n')
   dbcloseall()

   quit

   return (.f.)
